
__version__ = "0.0.0"

# __all__ = ['a', 'b', 'c']


import emma.analysis

from emma.objects import *
from emma.matrices import *

from emma.graphics import *
from emma.solvers import *

import emma.utilities
import emma.matlib


settings = {
    "DATAFRAME_LATEX": True,
}


